/////////////////////////////////////
//
// Gaussian Mixture Model
// written by tomoyuki nagahashi
//
/////////////////////////////////////

#include "GMM.h"

GMM::GMM() 
{
	m_Dim = 3;
	m_nGaussian = 5;
	m_isLearn = true;
}

GMM::~GMM() 
{
}


////////////////////
// initialization
////////////////////
int GMM::Init(int dim, int nGaussian)
{
	m_Dim = dim;
	m_nGaussian = nGaussian;
	m_Data.clear();
	m_Norm = pow(2.0*PI, m_Dim*0.5);
	m_isLearn = true;
	return 0;
}

////////////////////////////
// Gaussian Probability
////////////////////////////
double GMM::Prob(vector<double> &x)
{
	if(m_Data.size() == 0)
	{
		return 1;
	}
	double sum = 0.0;
	for(int i = 0; i < m_nGaussian; i++)
	{
		sum += Prob(x, i);
	}	
	return sum;
}

////////////////////////////
// Gaussian Probability
////////////////////////////
double GMM::Prob(vector<double> &x, int k)
{
	vector<double> diff(m_Dim);
	for(int i = 0; i < diff.size(); i++)
	{
		diff[i] = x[i] - m_M[k][i];
	}
	
	vector<double> tmp(m_Dim, 0.0);
	for(int i = 0; i < m_Dim; i++)
	{
		for(int j = 0; j < m_Dim; j++)
		{
			tmp[i] += diff[i]*m_InvS[k][j][i];
		}
	}
	double d = 0.0;
	for(int i = 0; i < m_Dim; i++)
	{
		d += tmp[i] * diff[i];
	}
	
	return exp(-d*0.5) / m_Norm / (m_Det[k]);
}



///////////////////////////////////////
// inverse covarince and determinant
///////////////////////////////////////
int GMM::CalcInvDet(void)
{
	m_InvS.clear();
	m_InvS.resize(m_nGaussian);
	m_Det.clear();
	m_Det.resize(m_nGaussian);
	for(int i = 0; i < m_nGaussian; i++)
	{
		m_InvS[i] = InverseMatrix( m_S[i] );
		m_Det[i] = sqrt( Det( m_S[i] ) );
 	}

	
	return 0;
}



////////////////////////
// inverse matrix
////////////////////////
vector< vector<double> > GMM::InverseMatrix(vector< vector<double> > &s)
{
	vector< vector<double> > a = s;
	vector< vector<double> > inv_a = a;
	for(int i = 0; i < inv_a.size(); i++)
	{
		for(int j = 0; j < inv_a[i].size(); j++)
		{
			inv_a[i][j] = 0.0;
		}
		inv_a[i][i] = 1.0;
	}
	
	double buf;
	for(int i = 0; i < inv_a.size(); i++)
	{
		buf = 1/a[i][i];
		for(int j = 0; j < inv_a[i].size(); j++)
		{
			a[i][j] *= buf;
			inv_a[i][j] *= buf;
		}
		for(int j = 0; j < inv_a[i].size(); j++)
		{
			if(i != j)
			{
				buf = a[j][i];
				for(int k = 0; k < a[j].size(); k++)
				{
					a[j][k] -= a[i][k]*buf;
					inv_a[j][k] -= inv_a[i][k]*buf;
				}
			}
		}
	}
	
	return inv_a;
}



////////////////////////
// Determinant
////////////////////////
double GMM::Det(vector< vector<double> > &s)
{
	vector< vector<double> > a = s;
	double buf;
	double det = 1.0;
	int size = s.size();
	
	for(int i = 0; i < size; i++)
	{
		for(int j = 0; j < size; j++)
		{
			if(i < j)
			{
				buf = a[j][i]/a[i][i];
				for(int k = 0; k < size; k++)
				{
					a[j][k] -= a[i][k]*buf;
				}
			}
		}
	}

	for(int i = 0; i < size; i++)
	{
		det *= a[i][i];
	}
	return det;
}



//////////////////
// set data
//////////////////
int GMM::SetData(vector<double> x)
{
	m_Data.push_back(x);
	return 0;
}

////////////////
// GMM learning
////////////////
int GMM::Learn()
{
	int datasize = m_Data.size();
	
	if(datasize == 0)
	{
		return 0;
	}
	
	m_Prob.clear();
	m_Prob.resize(datasize, 0.0);
	m_Post.clear();
	m_Post.resize(m_nGaussian, vector<double>(datasize, 0.0));
	m_Prior.clear();
	m_Prior.resize(m_nGaussian, vector<double>(datasize, 0.0));
	
	if(m_isLearn)
	{
		m_S.clear();
		m_S.resize(m_nGaussian, vector< vector<double> >(m_Dim, vector<double>(m_Dim, 0.0)));
		m_M.clear();
		m_M.resize(m_nGaussian, vector<double>(m_Dim, 0.0));
		m_Weight.clear();
		m_Weight.resize(m_nGaussian, 1.0/(double)m_nGaussian);

		//init 
		for(int i = 0; i < m_nGaussian; i++)
		{
			int n = rand()%m_Data.size();
			m_M[i] = m_Data[n];

			for(int j = 0; j < m_Dim; j++)
			{
				m_S[i][j][j] = 1.0;
			}	

		}
		m_isLearn = false;
	}

	CalcPx();
	
	// EM
	double likelihood = -1.0e+10;
	double diff_likelihood = 1.0e+10;
	int counter = 0;
	while( (diff_likelihood) > 1.0e-7)
	//while( counter < 10000)
	{
		double sum = 0.0;
		for(int i = 0; i < m_nGaussian; i++)
		{
			sum += m_Weight[i];
		}

		for(int i = 0; i < m_nGaussian; i++)
		{	
			EMProb(i);
			EMPar(i, m_Weight[i]/sum);
		}

		CalcPx();
		
		// clac likelihood
		double tmp = 0.0;
		for(int i = 0; i < m_Data.size(); i++)
		{
			if(m_Prob[i] > 0.0)
			{
				tmp += log(m_Prob[i]);
			}
		}
		diff_likelihood = likelihood;
		likelihood = tmp / (double)m_Data.size();
		diff_likelihood = fabs(diff_likelihood - likelihood);
		counter++;
		
		if(counter > 1000)
		{
			break;
		}
	}	
	cout << counter << ":" << diff_likelihood << endl;

	CalcInvDet();
	
	return 0;
}

////////////////////////
// prob all data
////////////////////////
int GMM::CalcPx(void)
{
	CalcInvDet();
	
	for(int i = 0; i < m_Data.size(); i++)
	{
		m_Prob[i] = 0;
		for(int j = 0; j < m_nGaussian; j++)
		{
			m_Prior[j][i] = Prob(m_Data[i], j);
			m_Prob[i] += m_Weight[j] * m_Prior[j][i];
		}
	}
	return 0;
}

/////////////////////////////////////////////////////
// calc posterior probability & gauss param->weight
/////////////////////////////////////////////////////
int GMM::EMProb(int nk)
{
	int np = m_Data.size();
	double value = 0.0;

	for(int j=0 ; j<np ; j++)
	{
		if(m_Prob[j] != 0)
		{
			m_Post[nk][j] = m_Weight[nk] * m_Prior[nk][j] / m_Prob[j];
		}
		else
		{
			m_Post[nk][j] = 0.0;
		}
		value += m_Post[nk][j];
	}

	if(value == 0.0){
		m_Weight[nk] = 1.0e-5;
	}
	else{
		m_Weight[nk] = value / (double)np;
	}

	return 0;
}



//////////////////////////////
// parameter update
//////////////////////////////
int GMM::EMPar(int nk, double prior)
{
	int i, j;
	int j1, j2;
	double psi;
	int np = m_Data.size();
	vector< vector<double> > tmpsx(m_Dim, vector<double>(m_Dim, 0.0));

	for(int i = 0; i < m_M[nk].size(); i++)
	{
		m_M[nk][i] = 0.0;
	}
	
	for(j = 0; j < np; j++) 
	{
		for(i = 0; i < m_Dim; i++)
		{
			m_M[nk][i] += m_Post[nk][j]* m_Data[j][i];
		}
		for(j1 = 0; j1 < m_Dim; j1++)
		{
			for(j2 = 0; j2 < m_Dim; j2++)
			{
				tmpsx[j1][j2] += m_Post[nk][j] * m_Data[j][j1]*m_Data[j][j2];
			}
		}
    }

	psi = 1.0 / ((double)np * m_Weight[nk]);
	
	// update mean
	for(i = 0; i < m_Dim; i++)
	{
		m_M[nk][i] *= psi;
	}
	// update conv matrix
	for(j1 = 0; j1 < m_Dim; j1++)
	{
		for(j2 = j1; j2 < m_Dim; j2++)
		{
			m_S[nk][j1][j2] = (tmpsx[j1][j2] * psi - m_M[nk][j1]*m_M[nk][j2]);
		}
	}
	
	m_Weight[nk] = 0.8 * m_Weight[nk] + 0.2 * prior;

	return 0;	
}